'''
Transgen module used to create transition graphs from grammars.
'''

import EasyExtend.fibers.grammar_fiber.fiber as grammar_fiber
symbol = grammar_fiber.symbol
token  = grammar_fiber.token
import EasyExtend.fibers.zero.fiber as zero_fiber
from EasyExtend.csttools import find_node, find_all
import os


def get_nid(item, fiber = zero_fiber):
    nid = fiber.symbol.__dict__.get(item)
    if nid is not None:
        return nid
    nid = fiber.token.TOKEN_MAP.get(item)
    if nid is not None:
        return nid
    else:
        return item

def gen_rule_table(fiber = zero_fiber):
    '''
    function used to create simple rule table from EBNF grammar description.
    @param fiber: optional fiber module. When available Grammar of fiber is parsed.
        Otherwise Pythons Grammar is used as a default.
    @return: dictionary of rules.
    '''
    grammar = grammar_fiber.load_fiber_grammar(fiber)
    rules = {}
    for r in find_all(grammar, symbol.RULE):
        rhs = find_node(r, symbol.RHS)
        rule_name = find_node(r, token.NAME)[1]
        nid = get_nid(rule_name, fiber)
        rule = FlatRule(fiber, nid)
        rule.create(rhs)
        rules[nid] = rule.flatten()
    return rules

def load_dfa_table(fiber = None):
    if not fiber:
        import EasyExtend.fibers.zero.eetransitions as pytransitions
        return pytransitions.rule_dfas
    else:
        mod_name = "EasyExtend"+".".join((fiber.__file__.split("EasyExtend")[1]).split(os.sep)[:-1])+".eetransitions"
        try:
            module = __import__(mod_name, globals(), locals(),["EasyExtend"])
            return module.rule_dfas
        except ImportError:
            rules = gen_rule_table(fiber)
            return create_transitions(fiber = fiber, rules = rules)


class FlatRule(object):
    def __init__(self, fiber = zero_fiber, nid=-1):
        self.nid   = nid
        self.ALT   = []
        self.ITEMS = []
        self.multiplicity = '1'
        self.fiber = fiber

    def create(self, node):
        alternatives = find_all(node, symbol.ALT, level = 1)
        if len(alternatives)>1:
            for alt in alternatives:
                rule = FlatRule(fiber = self.fiber)
                rule.ITEMS = self.from_alt(alt)
                self.ALT.append(rule)
        else:
            self.ITEMS = self.from_alt(alternatives[0])

    def from_alt(self, node):
        items = []
        for ITEM in find_all(node, symbol.ITEM, level = 1):
            ATOM = find_node(ITEM, symbol.ATOM, level = 1)
            if ATOM:
                rule = FlatRule(fiber = self.fiber)
                if find_node(ITEM, token.PLUS, level = 1):
                    rule.multiplicity = '+'
                if find_node(ITEM, token.STAR, level = 1):
                    rule.multiplicity = '*'
                rhs = find_node(ATOM, symbol.RHS)
                if rhs:
                    rule.create(rhs)
                else:
                    sym_name = find_node(ATOM, token.NAME)
                    if sym_name:
                        rule.nid = get_nid(sym_name[1], self.fiber)
                    else:
                        tok_name = find_node(ATOM, token.STRING)
                        rule.nid = get_nid(tok_name[1][1:-1], self.fiber)
                items.append(rule)
            else:
                rule = FlatRule(fiber = self.fiber)
                rule.multiplicity = '?'
                rule.create(find_node(ITEM, symbol.RHS))
                items.append(rule)
        return items

    def left_factor(self, alt_rule):
        if not alt_rule[0] == '|':
            return alt_rule
        rules = alt_rule[1:]
        typ = rules[0][0]
        prefix = []
        i = 0
        while True:
            try:
                c = rules[0][i]
                for r in rules[1:]:
                    if r[i]!=c:
                        break
                else:
                    i+=1
                    prefix.append(c)
                    continue
                break
            except IndexError:
                break
        if len(prefix)>1:
            m = prefix[0]
            prefix.append(['|']+[[m]+r[len(prefix):] for r in rules])
            return prefix
        else:
            return alt_rule

    def flatten(self):
        if self.ALT:
            _expect = ["|"]
            for r in self.ALT:
                if r.nid>=0:
                    _expect.append(r.nid)
                else:
                    _expect.append(r.flatten())
            return self.left_factor(_expect)
        elif self.ITEMS:
            _expect = [self.multiplicity]
            for r in self.ITEMS:
                if r.nid>=0:
                    if r.multiplicity == '*':
                        _expect.append(['*', r.nid])
                    elif r.multiplicity == '+':
                        _expect.append(r.nid)
                        _expect.append(['*',r.nid])
                    else:
                        _expect.append(r.nid)
                else:
                    flat = r.flatten()
                    if flat[0] == '+':
                        _expect.append(['1']+flat[1:])
                        _expect.append(['*']+flat[1:])
                    elif r.multiplicity in ('?','*') and r.multiplicity!=flat[0]:
                        _expect.append([r.multiplicity,flat])
                    else:
                        _expect.append(flat)
            if _expect[0] == '1' and len(_expect) == 2 and isinstance(_expect[1], list):
                return _expect[1]
            return _expect

EMPTY = (None, -1)

class GrammarRuleDFA(object):
    '''
    Assembles finite statemachines from rules.
    '''
    # A far to complicated implementation of grammar rule translations into finite automata.
    # I realized late that simplifications and disambiguations provided by the algorithm below
    # is not really required. GrammarTracer live well by stepping through different traces at
    # the same time.
    _cnt = 0
    def __init__(self, rules = {}):
        self.initial  = EMPTY
        self.terminal = EMPTY
        self.transitions = {}
        self._type = '1'  # used for concatenation only
        self.rules = rules

    @classmethod
    def create_all(cls, rules):
        grammar_dfas = {}
        for key, val in rules.items():
            #print key, val
            gr = GrammarRuleDFA(rules)
            f = gr.create_from(val)
            grammar_dfas[key] = [rules[key], f.initial, f.transitions]
        return grammar_dfas

    def create_from(self, lst, k = 0):
        if lst[0] == '|':
            return self.create_combine(lst, k)
        dfa = None
        dfas = []
        for item in lst[1:]:
            if isinstance(item, (int, str)):
                item = ['1',item]
                dfa = self.local_dfa(item,k)
            else:
                for sub in item:
                    if isinstance(sub, list):
                        dfa = self.create_from(item, k)
                        break
                else:
                    dfa = self.local_dfa(item,k)
            dfas.insert(0,dfa)
            k+=len(dfa.transitions)
        HH_connected = False
        while dfas:
            last_dfa = dfas[0]
            del dfas[0]
            for dfa in dfas:
                HH_connected = False
                dfa.connect_T_H(last_dfa)
                if dfa._type == '1':
                    break
                if dfa.connect_H_H(last_dfa):
                    HH_connected = True
                    break
        dfa.set_final()
        dfa._type = lst[0]
        if isinstance(lst[1], list) and lst[1][0] == '?' and not HH_connected:
            self.split_entry_point(dfa)
        if lst[0] == '*':
            dfa.transitions[dfa.terminal].append(dfa.initial)
        return dfa

    def split_entry_point(self, dfa):
        entry_node = ("entry_node",self._cnt)
        self._cnt+=1
        state_init = dfa.initial
        trans = dfa.transitions[state_init]
        dfa.initial = entry_node
        dfa.transitions[entry_node] = trans+[state_init]

    def set_final(self):
        '''
        Search one final node and set node to final.
        '''
        state = self.initial

        visited = [state]
        while 1:
            trans = self.transitions[state]
            if (None ,-1) in trans:
                self.terminal = state
                return
            else:
                for s in trans:
                    if s not in visited:
                        state = s
                        visited.append(s)
                        break

    def create_combine(self, lst, k):
        graphs = []
        for L in lst[1:]:
            g = self.create_from(L,k)
            graphs.append(g)
            k+=len(g.transitions)
        combine = GrammarRuleDFA(rules = self.rules)
        combine._type = '|'
        entry_node = ("entry_node",self._cnt)
        combine.initial = entry_node
        combine.transitions[entry_node] = [g.initial for g in graphs]
        terminal_node = ("term_node",self._cnt)
        self._cnt+=1
        terminal = GrammarRuleDFA(rules = self.rules)
        terminal.initial = terminal_node
        terminal.terminal = terminal_node
        terminal.transitions[terminal_node] = [EMPTY]
        for g in graphs:
            g.connect_T_H(terminal)
            combine.transitions.update(g.transitions)
        combine.terminal = terminal_node
        return combine

    def local_dfa(self, lst, k = 0):
        if lst[0] == '|':
            return self.create_combine(lst, k)
        dfa = GrammarRuleDFA(rules = self.rules)
        dfa._type = lst[0]
        state0 = (lst[1], k)
        dfa.initial = state0
        state = state0
        k+=1
        for item in lst[2:]:
            dfa.transitions[state] = [(item, k)]
            state = (item, k)
            k+=1
        dfa.terminal = state
        if lst[0] == '*':
            dfa.transitions[state] = [state0, EMPTY]
        else:
            dfa.transitions[state] = [EMPTY]
        return dfa

    def connect_T_H(self, other):
        if other.initial[0] == "entry_node":
            self.connect_Alt(other)
        final_transitions = self.transitions[self.terminal]
        if other._type == '1':
            if EMPTY in final_transitions:
                final_transitions.remove(EMPTY)
        if other.initial not in final_transitions:
            final_transitions.append(other.initial)
        self.transitions.update(other.transitions)

    def connect_Alt(H, T):
        if H._type in ('*','?'):
            dfa1, dfa2 = T.extract_trace(H.initial[0])
            if dfa1.initial != EMPTY:
                H.transitions[H.terminal].append(T.initial)
                H.connect_H_H(dfa2)
                H.connect_T_H(dfa1)
                return True
        return False

    def extract_trace(self, nid):
        def shift_states(states, dfa_new, dfa_old, visited = []):

            def filter(T):
                if T[0] == "term_node":
                    return EMPTY
                else:
                    return T

            for S in states:
                if S[0] == "term_node":
                    pass
                elif S!=EMPTY and S not in visited:
                    trans = dfa_old.transitions[S]
                    dfa_new.transitions[S] = [filter(T) for T in trans]
                    del dfa_old.transitions[S]
                    visited.append(S)
                    shift_states(trans, dfa_new, dfa_old, visited)

        if self.initial[0] == nid:
            return self, GrammarRuleDFA(rules = self.rules)
        elif self.initial[0] == "entry_node":
            for S in self.transitions[self.initial]:
                if S[0] == nid:
                    # start extracting trace
                    self.transitions[self.initial].remove(S)
                    dfa = GrammarRuleDFA(rules = self.rules)
                    dfa.initial = S
                    shift_states([S], dfa, self)
                    dfa.set_final()
                    return self, dfa
        return GrammarRuleDFA(rules = self.rules), self

    def connect_H_H(self, other):
        def cleanup(states):
            _states = sorted(states)
            last = _states[-1]
            for s in _states[-2::-1]:
                if s[0] == last[0]:
                    states.remove(last)
                    prev_trans = self.transitions[last]
                    del self.transitions[last]
                    this_trans = self.transitions[s]
                    for item in prev_trans:
                        if item not in this_trans:
                            this_trans.append(item)
                    for V in self.transitions.values():
                        if last in V:
                            V.remove(last)
                            V.append(s)
                    cleanup(this_trans)
                last = s

        def merge(H,T):
            initial_transitions = H.transitions[H.initial]
            initial_transitions+= T.transitions[T.initial]
            H.transitions.update(T.transitions)
            if T._type=='1':
                for states in H.transitions.values():
                    cleanup(states)
                return True
            del H.transitions[T.initial]
            # replace other.initial by self.initial
            collected = {}
            double = None
            for states in H.transitions.values():
                if T.initial in states:
                    states.remove(T.initial)
                    if H.initial not in states:
                        states.append(H.initial)
                cleanup(states)
            return True

        if other.initial[0] == self.initial[0]:
            return merge(self, other)

        return False

    def __str__(self):
        return str(self.transitions)


class NonSelectableError(Exception):pass

class InvalidNodeError(Exception):pass

class GrammarTracer:
    def __init__(self, fiber = zero_fiber):
        self.dfa_table  = load_dfa_table(fiber)

class RuleTracer:
    def __init__(self, table):
        self.dfa_table  = table
        self.state      = None   # can be shared by more than one node with same nid
        self.selectable = []
        self.dfa = None

    def clone(self):
        rt = self.__class__(self.dfa_table)
        rt.selectable = self.selectable
        rt.state = self.state
        rt.dfa   = self.dfa
        return rt

    def get_nid(self, state):
        return state[0][0]

    def select(self, nid):
        try:
            if self.state is None:
                self.dfa   = self.dfa_table[nid][2]
                self.state = [self.dfa_table[nid][1]]
                if self.state[0][0] != "entry_node":
                    self.selectable = self.state
                    return [self.get_nid(self.state)]
                return self._next()
            if nid is None:
                raise NonSelectableError(str(nid))
            self.state = [item for item in self.selectable if item[0] == nid]
            if self.state:
                return self._next()
            raise NonSelectableError(str(nid))
        except KeyError:
            raise NonSelectableError(str(nid))

    def _get_next_state(self, state):
        states = []
        if len(state) == 1:
            trans = self.dfa[state[0]]
            for item in trans:
                if item[0] in ("entry_node", "term_node"):
                    states+=self._get_next_state([item])
                else:
                    states.append(item)
        else:
            for S in state:
                states.extend(self._get_next_state([S]))
        return states

    def _next(self):
        if self.state is [None]:
            raise StopIteration
        self.selectable = self._get_next_state(self.state)
        selections = []
        for s in self.selectable:
            if s[0] not in selections:
                selections.append(s[0])
        return selections

    def check_node(self, node):
        '''
        checks validity of a CST node against a grammar.

        Note: This implementation is currently able to check against the Python grammar only.
        '''
        nid = node[0]%512
        next = self.select(nid)
        subnodes = []
        for sub in node[1:]:
            nid = sub[0]%512
            if isinstance(sub, list):
                if nid == 1 and nid not in next and sub[1] not in next:
                    raise InvalidNodeError(str(sub))
                elif nid in next:
                    if nid >= 256:
                        self.check_node(sub)
                    next = self.select(nid)
                else:
                    raise InvalidNodeError(str(sub))
        if None not in next:
            raise InvalidNodeError(str(node))
        return True


def create_transitions(fiber = zero_fiber, rules = {}):
    from EasyExtend.util.path import path
    fPyTrans = open(path(fiber.__file__).dirname().joinpath("eetransitions.py"),"w")
    print "*** Modify %s ***"%fPyTrans.name
    print >> fPyTrans, "# %s" % ("_" * 70)
    print >> fPyTrans, "# This was automatically generated by eetransgen.py."
    print >> fPyTrans, "# Hack at your own risk."
    print >> fPyTrans
    if not rules:
        rules = gen_rule_table(fiber)
    grammar_dfas = GrammarRuleDFA.create_all(rules)
    import pprint
    print >> fPyTrans, "rule_dfas = "+pprint.pformat(grammar_dfas)
    fPyTrans.close()
    return grammar_dfas


if __name__ == '__main__':
    create_transitions()